@@ -13,7 +13,6 @@
 -- License:     BSD3
 -- Maintainer:  Leon P Smith <leon@melding-monads.com>
 -- Stability:   experimental
--- Portability: portable
 --
 -- A mid-level client library for the PostgreSQL database, aimed at ease of
 -- use and high performance.
@@ -67,6 +66,7 @@ module Database.PostgreSQL.Simple
     , (:.)(..)
     -- ** Exceptions
     , SqlError(..)
+    , PQ.ExecStatus(..)
     , FormatError(fmtMessage, fmtQuery, fmtParams)
     , QueryError(qeMessage, qeQuery)
     , ResultError(errSQLType, errHaskellType, errMessage)
@@ -97,20 +97,9 @@ module Database.PostgreSQL.Simple
 --    , Base.insertID
     -- * Transaction handling
     , withTransaction
-    , withTransactionSerializable
-    , TransactionMode(..)
-    , IsolationLevel(..)
-    , ReadWriteMode(..)
-    , defaultTransactionMode
-    , defaultIsolationLevel
-    , defaultReadWriteMode
-    , withTransactionLevel
-    , withTransactionMode
-    , withTransactionModeRetry
+    , withSavepoint
 --    , Base.autocommit
     , begin
-    , beginLevel
-    , beginMode
     , commit
     , rollback
     -- * Helper functions
@@ -121,20 +110,16 @@ module Database.PostgreSQL.Simple
 import           Blaze.ByteString.Builder
                    ( Builder, fromByteString, toByteString )
 import           Blaze.ByteString.Builder.Char8 (fromChar)
+import           Blaze.Text ( integral )
 import           Control.Applicative ((<$>), pure)
-import           Control.Concurrent.MVar
-import           Control.Exception
-                   ( Exception, onException, throw, throwIO, finally
-                   , try, SomeException, fromException )
+import           Control.Exception as E
 import           Control.Monad (foldM)
 import           Data.ByteString (ByteString)
 import           Data.Int (Int64)
-import qualified Data.IntMap as IntMap
 import           Data.List (intersperse)
-import           Data.Monoid (mappend, mconcat)
+import           Data.Monoid (mconcat)
 import           Data.Typeable (Typeable)
-import           Database.PostgreSQL.Simple.BuiltinTypes ( oid2typname )
-import           Database.PostgreSQL.Simple.Compat ( mask )
+import           Database.PostgreSQL.Simple.Compat ( (<>) )
 import           Database.PostgreSQL.Simple.FromField (ResultError(..))
 import           Database.PostgreSQL.Simple.FromRow (FromRow(..))
 import           Database.PostgreSQL.Simple.Ok
@@ -143,11 +128,12 @@ import           Database.PostgreSQL.Simple.ToRow (ToRow(..))
 import           Database.PostgreSQL.Simple.Types
                    ( Binary(..), In(..), Only(..), Query(..), (:.)(..) )
 import           Database.PostgreSQL.Simple.Internal as Base
+import           Database.PostgreSQL.Simple.Transaction
+import           Database.PostgreSQL.Simple.TypeInfo
 import qualified Database.PostgreSQL.LibPQ as PQ
 import qualified Data.ByteString.Char8 as B
 import qualified Data.Text          as T
 import qualified Data.Text.Encoding as TE
-import qualified Data.Vector as V
 import           Control.Monad.Trans.Reader
 import           Control.Monad.Trans.State.Strict
 
@@ -322,7 +308,7 @@ buildQuery conn q template xs = zipParams (split template) <$> mapM sub xs
         sub (Many  ys)      = mconcat <$> mapM sub ys
         split s = fromByteString h : if B.null t then [] else split (B.tail t)
             where (h,t) = B.break (=='?') s
-        zipParams (t:ts) (p:ps) = t `mappend` p `mappend` zipParams ts ps
+        zipParams (t:ts) (p:ps) = t <> p <> zipParams ts ps
         zipParams [t] []        = t
         zipParams _ _ = fmtError (show (B.count '?' template) ++
                                   " '?' characters, but " ++
@@ -476,7 +462,7 @@ doFold :: ( FromRow row )
        -> a
        -> (a -> row -> IO a)
        -> IO a
-doFold FoldOptions{..} conn _template q a f = do
+doFold FoldOptions{..} conn _template q a0 f = do
     stat <- withConnection conn PQ.transactionStatus
     case stat of
       PQ.TransIdle    -> withTransactionMode transactionMode conn go
@@ -494,11 +480,28 @@ doFold FoldOptions{..} conn _template q a f = do
       PQ.TransUnknown -> fail "foldWithOpts FIXME:  PQ.TransUnknown"
          -- Not sure what this means.
   where
-    go = do
-       -- FIXME:  what about name clashes with already-declared cursors?
-       _ <- execute_ conn ("DECLARE fold NO SCROLL CURSOR FOR "
-                           `mappend` q)
-       loop a `finally` execute_ conn "CLOSE fold"
+    declare = do
+        name <- newTempName conn
+        _ <- execute_ conn $ mconcat
+                 [ "DECLARE ", name, " NO SCROLL CURSOR FOR ", q ]
+        return name
+    fetch (Query name) = query_ conn $
+        Query (toByteString (fromByteString "FETCH FORWARD "
+                             <> integral chunkSize
+                             <> fromByteString " FROM "
+                             <> fromByteString name
+                            ))
+    close name =
+        (execute_ conn ("CLOSE " <> name) >> return ()) `E.catch` \ex ->
+            -- Don't throw exception if CLOSE failed because the transaction is
+            -- aborted.  Otherwise, it will throw away the original error.
+            if isFailedTransactionError ex then return () else throwIO ex
+
+    go = bracket declare close $ \name ->
+         let loop a = do
+                 rs <- fetch name
+                 if null rs then return a else foldM f a rs >>= loop
+          in loop a0
 
 -- FIXME: choose the Automatic chunkSize more intelligently
 --   One possibility is to use the type of the results,  although this
@@ -509,9 +512,6 @@ doFold FoldOptions{..} conn _template q a f = do
     chunkSize = case fetchQuantity of
                  Automatic   -> 256
                  Fixed n     -> n
-    loop a = do
-      rs <- query conn "FETCH FORWARD ? FROM fold" (Only chunkSize)
-      if null rs then return a else foldM f a rs >>= loop
 
 -- | A version of 'fold' that does not transform a state value.
 forEach :: (ToRow q, FromRow r) =>
@@ -550,28 +550,28 @@ finishQuery conn q result = do
     PQ.CommandOk -> do
         throwIO $ QueryError "query resulted in a command response" q
     PQ.TuplesOk -> do
-        ncols <- PQ.nfields result
         let unCol (PQ.Col x) = fromIntegral x :: Int
-        typenames <- V.generateM (unCol ncols)
-                                 (\(PQ.Col . fromIntegral -> col) -> do
-                                    getTypename conn =<< PQ.ftype result col)
         nrows <- PQ.ntuples result
         ncols <- PQ.nfields result
         forM' 0 (nrows-1) $ \row -> do
-           let rw = Row row typenames result
-           case runStateT (runReaderT (unRP fromRow) rw) 0 of
+           let rw = Row row result
+           okvc <- runConversion (runStateT (runReaderT (unRP fromRow) rw) 0) conn
+           case okvc of
              Ok (val,col) | col == ncols -> return val
                           | otherwise -> do
                               vals <- forM' 0 (ncols-1) $ \c -> do
+                                  tinfo <- getTypeInfo conn =<< PQ.ftype result c
                                   v <- PQ.getvalue result row c
-                                  return ( typenames V.! unCol c
+                                  return ( tinfo
                                          , fmap ellipsis v       )
                               throw (ConversionFailed
                                (show (unCol ncols) ++ " values: " ++ show vals)
+                               Nothing
+                               ""
                                (show (unCol col) ++ " slots in target type")
                                "mismatch between number of columns to \
                                \convert and number in target type")
-             Errors []  -> throwIO $ ConversionFailed "" "" "unknown error"
+             Errors []  -> throwIO $ ConversionFailed "" Nothing "" "" "unknown error"
              Errors [x] -> throwIO x
              Errors xs  -> throwIO $ ManyErrors xs
     PQ.CopyOut ->
@@ -587,160 +587,6 @@ ellipsis bs
     | B.length bs > 15 = B.take 10 bs `B.append` "[...]"
     | otherwise        = bs
 
-
--- | Of the four isolation levels defined by the SQL standard,
--- these are the three levels distinguished by PostgreSQL as of version 9.0.
--- See <http://www.postgresql.org/docs/9.1/static/transaction-iso.html>
--- for more information.   Note that prior to PostgreSQL 9.0, 'RepeatableRead'
--- was equivalent to 'Serializable'.
-
-data IsolationLevel
-   = DefaultIsolationLevel  -- ^ the isolation level will be taken from
-                            --   PostgreSQL's per-connection
-                            --   @default_transaction_isolation@ variable,
-                            --   which is initialized according to the
-                            --   server's config.  The default configuration
-                            --   is 'ReadCommitted'.
-   | ReadCommitted
-   | RepeatableRead
-   | Serializable
-     deriving (Show, Eq, Ord, Enum, Bounded)
-
-data ReadWriteMode
-   = DefaultReadWriteMode   -- ^ the read-write mode will be taken from
-                            --   PostgreSQL's per-connection
-                            --   @default_transaction_read_only@ variable,
-                            --   which is initialized according to the
-                            --   server's config.  The default configuration
-                            --   is 'ReadWrite'.
-   | ReadWrite
-   | ReadOnly
-     deriving (Show, Eq, Ord, Enum, Bounded)
-
-data TransactionMode = TransactionMode {
-       isolationLevel :: !IsolationLevel,
-       readWriteMode  :: !ReadWriteMode
-     } deriving (Show, Eq)
-
-defaultTransactionMode :: TransactionMode
-defaultTransactionMode =  TransactionMode
-                            defaultIsolationLevel
-                            defaultReadWriteMode
-
-defaultIsolationLevel  :: IsolationLevel
-defaultIsolationLevel  =  DefaultIsolationLevel
-
-defaultReadWriteMode   :: ReadWriteMode
-defaultReadWriteMode   =  DefaultReadWriteMode
-
--- | Execute an action inside a SQL transaction.
---
--- This function initiates a transaction with a \"@begin
--- transaction@\" statement, then executes the supplied action.  If
--- the action succeeds, the transaction will be completed with
--- 'Base.commit' before this function returns.
---
--- If the action throws /any/ kind of exception (not just a
--- PostgreSQL-related exception), the transaction will be rolled back using
--- 'rollback', then the exception will be rethrown.
-withTransaction :: Connection -> IO a -> IO a
-withTransaction = withTransactionMode defaultTransactionMode
-
--- | Execute an action inside of a 'Serializable' transaction.  If a
--- serialization failure occurs, roll back the transaction and try again.
--- Be warned that this may execute the IO action multiple times.
---
--- A 'Serializable' transaction creates the illusion that your program has
--- exclusive access to the database.  This means that, even in a concurrent
--- setting, you can perform queries in sequence without having to worry about
--- what might happen between one statement and the next.
---
--- Think of it as STM, but without @retry@.
-withTransactionSerializable :: Connection -> IO a -> IO a
-withTransactionSerializable =
-    withTransactionModeRetry
-        TransactionMode
-        { isolationLevel = Serializable
-        , readWriteMode  = ReadWrite
-        }
-
--- | Execute an action inside a SQL transaction with a given isolation level.
-withTransactionLevel :: IsolationLevel -> Connection -> IO a -> IO a
-withTransactionLevel lvl
-    = withTransactionMode defaultTransactionMode { isolationLevel = lvl }
-
--- | Execute an action inside a SQL transaction with a given transaction mode.
-withTransactionMode :: TransactionMode -> Connection -> IO a -> IO a
-withTransactionMode mode conn act =
-  mask $ \restore -> do
-    beginMode mode conn
-    r <- restore act `onException` rollback conn
-    commit conn
-    return r
-
--- | Like 'withTransactionMode', but if a 'SqlError' arises whose 'sqlState' is
--- @\"40001\"@ (@serialization_failure@), this will issue a @ROLLBACK@, then
--- try the action again.  If any other exception arises, this will issue a
--- @ROLLBACK@, but will propagate the exception instead of retrying.
---
--- This is used to implement 'withTransactionSerializable'.
-withTransactionModeRetry :: TransactionMode -> Connection -> IO a -> IO a
-withTransactionModeRetry mode conn act =
-    mask $ \restore ->
-        retryLoop $ try $ do
-            a <- restore act
-            commit conn
-            return a
-  where
-    retryLoop :: IO (Either SomeException a) -> IO a
-    retryLoop act' = do
-        beginMode mode conn
-        r <- act'
-        case r of
-            Left e -> do
-                rollback conn
-                case fromException e of
-                    Just SqlError{..} | sqlState == serialization_failure
-                      -> retryLoop act'
-                    _ -> throwIO e
-            Right a ->
-                return a
-
-    -- http://www.postgresql.org/docs/current/static/errcodes-appendix.html
-    serialization_failure = "40001"
-
--- | Rollback a transaction.
-rollback :: Connection -> IO ()
-rollback conn = execute_ conn "ABORT" >> return ()
-
--- | Commit a transaction.
-commit :: Connection -> IO ()
-commit conn = execute_ conn "COMMIT" >> return ()
-
--- | Begin a transaction.
-begin :: Connection -> IO ()
-begin = beginMode defaultTransactionMode
-
--- | Begin a transaction with a given isolation level
-beginLevel :: IsolationLevel -> Connection -> IO ()
-beginLevel lvl = beginMode defaultTransactionMode { isolationLevel = lvl }
-
--- | Begin a transaction with a given transaction mode
-beginMode :: TransactionMode -> Connection -> IO ()
-beginMode mode conn = do
-    _ <- execute_ conn $! Query (B.concat ["BEGIN", isolevel, readmode])
-    return ()
-  where
-    isolevel = case isolationLevel mode of
-                 DefaultIsolationLevel -> ""
-                 ReadCommitted  -> " ISOLATION LEVEL READ COMMITTED"
-                 RepeatableRead -> " ISOLATION LEVEL REPEATABLE READ"
-                 Serializable   -> " ISOLATION LEVEL SERIALIZABLE"
-    readmode = case readWriteMode mode of
-                 DefaultReadWriteMode -> ""
-                 ReadWrite -> " READ WRITE"
-                 ReadOnly  -> " READ ONLY"
-
 fmtError :: String -> Query -> [Action] -> a
 fmtError msg q xs = throw FormatError {
                       fmtMessage = msg
@@ -849,7 +695,7 @@ fmtError msg q xs = throw FormatError {
 --
 -- The obvious approach would appear to be something like this:
 --
--- > instance (Param a) => QueryParam a where
+-- > instance (ToField a) => ToRow a where
 -- >     ...
 --
 -- Unfortunately, this wreaks havoc with type inference, so we take a
@@ -905,7 +751,7 @@ fmtError msg q xs = throw FormatError {
 -- * There must be no other \"@?@\" characters anywhere in your
 --   template.
 --
--- * There must one or more \"@?@\" in the parentheses.
+-- * There must be one or more \"@?@\" in the parentheses.
 --
 -- * Extra white space is fine.
 --
@@ -928,9 +774,10 @@ fmtError msg q xs = throw FormatError {
 
 -- $returning
 --
--- The 'returning' function is similar to 'executeMany' but is
--- intended for use with multi-tuple @INSERT@ or @UPDATE@ statements
--- that make use of @RETURNING@.
+-- PostgreSQL supports returning values from data manipulation statements
+-- such as @INSERT@ and @UPDATE@.   You can use these statements by
+-- using 'query' instead of 'execute'.   For multi-tuple inserts,
+-- use 'returning' instead of 'executeMany'.
 --
 -- For example, were there an auto-incrementing @id@ column and
 -- timestamp column @t@ that defaulted to the present time for the
@@ -938,7 +785,8 @@ fmtError msg q xs = throw FormatError {
 -- sales records and also return their new @id@s and timestamps.
 --
 -- > let q = "insert into sales (amount, label) values (?,?) returning id, t"
--- > xs :: [(Int, UTCTime)] <- returning conn q [(20,"Chips"),(300,"Wood")]
+-- > xs :: [(Int, UTCTime)] <- query conn q (15,"Sawdust")
+-- > ys :: [(Int, UTCTime)] <- returning conn q [(20,"Chips"),(300,"Wood")]
 
 -- $result
 --
@@ -974,7 +822,7 @@ fmtError msg q xs = throw FormatError {
 --
 -- Although SQL can accommodate @NULL@ as a value for any of these
 -- types, Haskell cannot. If your result contains columns that may be
--- @NULL@, be sure that you use 'Maybe' in those positions of of your
+-- @NULL@, be sure that you use 'Maybe' in those positions of your
 -- tuple.
 --
 -- > (Text, Maybe Int, Int)
@@ -1014,25 +862,3 @@ fmtError msg q xs = throw FormatError {
 --   UTF-8. If you use some other encoding, decoding may fail or give
 --   wrong results. In such cases, write a @newtype@ wrapper and a
 --   custom 'Result' instance to handle your encoding.
-
-getTypename :: Connection -> PQ.Oid -> IO ByteString
-getTypename conn@Connection{..} oid =
-  case oid2typname oid of
-    Just name -> return name
-    Nothing -> modifyMVar connectionObjects $ \oidmap -> do
-      case IntMap.lookup (oid2int oid) oidmap of
-        Just name -> return (oidmap, name)
-        Nothing -> do
-            names <- query conn "SELECT typname FROM pg_type WHERE oid=?"
-                            (Only oid)
-            name <- case names of
-                      []  -> return $ throw SqlError {
-                                         sqlNativeError = -1,
-                                         sqlErrorMsg    = "invalid type oid",
-                                         sqlState       = ""
-                                       }
-                      [Only x] -> return x
-                      _   -> fail "typename query returned more than one result"
-                               -- oid is a primary key,  so the query should
-                               -- never return more than one result
-            return (IntMap.insert (oid2int oid) name oidmap, name)
