@@ -7,7 +7,6 @@
 -- License:     BSD3
 -- Maintainer:  Leon P Smith <leon@melding-monads.com>
 -- Stability:   experimental
--- Portability: portable
 --
 -- The 'FromRow' typeclass, for converting a row of results
 -- returned by a SQL query into a more useful Haskell representation.
@@ -25,27 +24,26 @@ module Database.PostgreSQL.Simple.FromRow
      ) where
 
 import Control.Applicative (Applicative(..), (<$>))
-import Control.Exception (SomeException(..))
 import Control.Monad (replicateM)
 import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as B
 import Database.PostgreSQL.Simple.Types (Only(..))
-import Database.PostgreSQL.Simple.Ok
 import qualified Database.PostgreSQL.LibPQ as PQ
 import           Database.PostgreSQL.Simple.Internal
+import           Database.PostgreSQL.Simple.Compat
 import           Database.PostgreSQL.Simple.FromField
+import           Database.PostgreSQL.Simple.Ok
 import           Database.PostgreSQL.Simple.Types ((:.)(..))
+import           Database.PostgreSQL.Simple.TypeInfo
 
 import Control.Monad.Trans.State.Strict
 import Control.Monad.Trans.Reader
 import Control.Monad.Trans.Class
 
-import Data.Vector ((!))
-
 -- | A collection type that can be converted from a sequence of fields.
 -- Instances are provided for tuples up to 10 elements and lists of any length.
 --
--- Note that instances can defined outside of postgresql-simple,  which is
+-- Note that instances can be defined outside of postgresql-simple,  which is
 -- often useful.   For example, here's an instance for a user-defined pair:
 --
 -- @data User = User { name :: String, fileQuota :: Int }
@@ -65,30 +63,45 @@ import Data.Vector ((!))
 class FromRow a where
     fromRow :: RowParser a
 
+getvalue :: PQ.Result -> PQ.Row -> PQ.Column -> Maybe ByteString
+getvalue result row col = unsafeDupablePerformIO (PQ.getvalue result row col)
+
+nfields :: PQ.Result -> PQ.Column
+nfields result = unsafeDupablePerformIO (PQ.nfields result)
+
+getTypeInfoByCol :: Row -> PQ.Column -> Conversion TypeInfo
+getTypeInfoByCol Row{..} col =
+    Conversion $ \conn -> do
+      oid <- PQ.ftype rowresult col
+      Ok <$> getTypeInfo conn oid
+
+getTypenameByCol :: Row -> PQ.Column -> Conversion ByteString
+getTypenameByCol row col = typname <$> getTypeInfoByCol row col
+
 fieldWith :: FieldParser a -> RowParser a
 fieldWith fieldP = RP $ do
     let unCol (PQ.Col x) = fromIntegral x :: Int
-    Row{..} <- ask
+    r@Row{..} <- ask
     column <- lift get
     lift (put (column + 1))
     let ncols = nfields rowresult
     if (column >= ncols)
-    then do
-        let vals = map (\c -> ( typenames ! (unCol c)
-                              , fmap ellipsis (getvalue rowresult row c) ))
-                       [0..ncols-1]
-            convertError = ConversionFailed
-                (show (unCol ncols) ++ " values: " ++ show vals)
+    then lift $ lift $ do
+        vals <- mapM (getTypenameByCol r) [0..ncols-1]
+        let err = ConversionFailed
+                (show (unCol ncols) ++ " values: " ++ show (map ellipsis vals))
+                Nothing
+                ""
                 ("at least " ++ show (unCol column + 1)
                   ++ " slots in target type")
                 "mismatch between number of columns to \
                 \convert and number in target type"
-        lift (lift (Errors [SomeException convertError]))
+        conversionError err
     else do
-        let typename = typenames ! unCol column
-            result = rowresult
-            field = Field{..}
-        lift (lift (fieldP field (getvalue result row column)))
+      let !result = rowresult
+          !typeOid = unsafeDupablePerformIO (PQ.ftype result column)
+          !field = Field{..}
+      lift (lift (fieldP field (getvalue result row column)))
 
 field :: FromField a => RowParser a
 field = fieldWith fromField
