@@ -1,4 +1,8 @@
-{-# LANGUAGE RecordWildCards, DeriveDataTypeable, GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE RecordWildCards #-}
+
 ------------------------------------------------------------------------------
 -- |
 -- Module:      Database.PostgreSQL.Simple.Internal
@@ -6,7 +10,6 @@
 -- License:     BSD3
 -- Maintainer:  Leon P Smith <leon@melding-monads.com>
 -- Stability:   experimental
--- Portability: portable
 --
 -- Internal bits.  This interface is less stable and can change at any time.
 -- In particular this means that while the rest of the postgresql-simple
@@ -19,29 +22,29 @@
 
 module Database.PostgreSQL.Simple.Internal where
 
-import Prelude hiding (catch)
-
 import           Control.Applicative
 import           Control.Exception
 import           Control.Concurrent.MVar
+import           Control.Monad(MonadPlus(..))
 import           Data.ByteString(ByteString)
-import qualified Data.ByteString       as B
 import qualified Data.ByteString.Char8 as B8
 import           Data.Char (ord)
 import           Data.Int (Int64)
 import qualified Data.IntMap as IntMap
+import           Data.IORef
+import           Data.Maybe(fromMaybe)
 import           Data.String
 import           Data.Typeable
 import           Data.Word
 import           Database.PostgreSQL.LibPQ(Oid(..))
 import qualified Database.PostgreSQL.LibPQ as PQ
-import           Database.PostgreSQL.Simple.BuiltinTypes (BuiltinType)
+import           Database.PostgreSQL.LibPQ(ExecStatus(..))
 import           Database.PostgreSQL.Simple.Ok
 import           Database.PostgreSQL.Simple.Types (Query(..))
+import           Database.PostgreSQL.Simple.TypeInfo.Types(TypeInfo)
 import           Control.Monad.Trans.State.Strict
 import           Control.Monad.Trans.Reader
-import qualified Data.Vector as V
-import           System.IO.Unsafe (unsafePerformIO)
+import           GHC.IO.Exception
 
 -- | A Field represents metadata about a particular field
 --
@@ -52,43 +55,30 @@ import           System.IO.Unsafe (unsafePerformIO)
 data Field = Field {
      result   :: !PQ.Result
    , column   :: {-# UNPACK #-} !PQ.Column
-   , typename :: !ByteString
+   , typeOid  :: {-# UNPACK #-} !PQ.Oid
+     -- ^ This returns the type oid associated with the column.  Analogous
+     --   to libpq's @PQftype@.
    }
 
-name :: Field -> Maybe ByteString
-name Field{..} = unsafePerformIO (PQ.fname result column)
-
-tableOid :: Field -> PQ.Oid
-tableOid Field{..} = unsafePerformIO (PQ.ftable result column)
-
-tableColumn :: Field -> Int
-tableColumn Field{..} = fromCol (unsafePerformIO (PQ.ftablecol result column))
-  where
-    fromCol (PQ.Col x) = fromIntegral x
-
-
-format :: Field -> PQ.Format
-format Field{..} = unsafePerformIO (PQ.fformat result column)
-
-typeOid :: Field -> PQ.Oid
-typeOid Field{..} = unsafePerformIO (PQ.ftype result column)
-
+type TypeInfoCache = IntMap.IntMap TypeInfo
 
 data Connection = Connection {
      connectionHandle  :: {-# UNPACK #-} !(MVar PQ.Connection)
-   , connectionObjects :: {-# UNPACK #-} !(MVar (IntMap.IntMap ByteString))
+   , connectionObjects :: {-# UNPACK #-} !(MVar TypeInfoCache)
+   , connectionTempNameCounter :: {-# UNPACK #-} !(IORef Int64)
    }
 
-data SqlType
-   = Builtin BuiltinType
-   | Other   Oid
-
 data SqlError = SqlError {
      sqlState       :: ByteString
-   , sqlNativeError :: Int
+   , sqlExecStatus  :: ExecStatus
    , sqlErrorMsg    :: ByteString
+   , sqlErrorDetail :: ByteString
+   , sqlErrorHint   :: ByteString
    } deriving (Show, Typeable)
 
+fatalError :: ByteString -> SqlError
+fatalError msg = SqlError "" FatalError msg "" ""
+
 instance Exception SqlError
 
 -- | Exception thrown if 'query' is used to perform an @INSERT@-like
@@ -152,6 +142,7 @@ connectPostgreSQL connstr = do
       PQ.ConnectionOk -> do
           connectionHandle  <- newMVar conn
           connectionObjects <- newMVar (IntMap.empty)
+          connectionTempNameCounter <- newIORef 0
           let wconn = Connection{..}
           version <- PQ.serverVersion conn
           let settings
@@ -162,9 +153,7 @@ connectPostgreSQL connstr = do
           return wconn
       _ -> do
           msg <- maybe "connectPostgreSQL error" id <$> PQ.errorMessage conn
-          throwIO $ SqlError { sqlNativeError = -1   -- FIXME?
-                             , sqlErrorMsg    = msg
-                             , sqlState       = ""  }
+          throwIO $ fatalError msg
 
 -- | Turns a 'ConnectInfo' data structure into a libpq connection string.
 
@@ -213,9 +202,7 @@ exec conn sql =
         case mres of
           Nothing -> do
             msg <- maybe "execute error" id <$> PQ.errorMessage h
-            throwIO $ SqlError { sqlNativeError = -1   -- FIXME?
-                               , sqlErrorMsg    = msg
-                               , sqlState       = ""  }
+            throwIO $ fatalError msg
           Just res -> do
             return res
 
@@ -260,21 +247,22 @@ finishExecute _conn q result = do
                     else error ("finishExecute:  not an int: " ++ B8.unpack str)
 
 throwResultError :: ByteString -> PQ.Result -> PQ.ExecStatus -> IO a
-throwResultError context result status = do
-    errormsg  <- maybe "" id <$> PQ.resultErrorMessage result
-    statusmsg <- PQ.resStatus status
+throwResultError _ result status = do
+    errormsg  <- fromMaybe "" <$>
+                 PQ.resultErrorField result PQ.DiagMessagePrimary
+    detail    <- fromMaybe "" <$>
+                 PQ.resultErrorField result PQ.DiagMessageDetail
+    hint      <- fromMaybe "" <$>
+                 PQ.resultErrorField result PQ.DiagMessageHint
     state     <- maybe "" id <$> PQ.resultErrorField result PQ.DiagSqlstate
     throwIO $ SqlError { sqlState = state
-                       , sqlNativeError = fromEnum status
-                       , sqlErrorMsg = B.concat [ context, ": ", statusmsg
-                                                , ": ", errormsg ]}
+                       , sqlExecStatus = status
+                       , sqlErrorMsg = errormsg
+                       , sqlErrorDetail = detail
+                       , sqlErrorHint = hint }
 
 disconnectedError :: SqlError
-disconnectedError = SqlError {
-                      sqlNativeError = -1,
-                      sqlErrorMsg    = "connection disconnected",
-                      sqlState       = ""
-                    }
+disconnectedError = fatalError "connection disconnected"
 
 -- | Atomically perform an action with the database handle, if there is one.
 withConnection :: Connection -> (PQ.Connection -> IO a) -> IO a
@@ -297,19 +285,69 @@ newNullConnection :: IO Connection
 newNullConnection = do
     connectionHandle  <- newMVar =<< PQ.newNullConnection
     connectionObjects <- newMVar IntMap.empty
+    connectionTempNameCounter <- newIORef 0
     return Connection{..}
 
 data Row = Row {
      row        :: {-# UNPACK #-} !PQ.Row
-   , typenames  :: !(V.Vector ByteString)
    , rowresult  :: !PQ.Result
    }
 
-newtype RowParser a = RP { unRP :: ReaderT Row (StateT PQ.Column Ok) a }
+newtype RowParser a = RP { unRP :: ReaderT Row (StateT PQ.Column Conversion) a }
    deriving ( Functor, Applicative, Alternative, Monad )
 
-getvalue :: PQ.Result -> PQ.Row -> PQ.Column -> Maybe ByteString
-getvalue result row col = unsafePerformIO (PQ.getvalue result row col)
+newtype Conversion a = Conversion { runConversion :: Connection -> IO (Ok a) }
 
-nfields :: PQ.Result -> PQ.Column
-nfields result = unsafePerformIO (PQ.nfields result)
+instance Functor Conversion where
+   fmap f m = Conversion $ \conn -> (fmap . fmap) f (runConversion m conn)
+
+instance Applicative Conversion where
+   pure a    = Conversion $ \_conn -> pure (pure a)
+   mf <*> ma = Conversion $ \conn -> do
+                   okf <- runConversion mf conn
+                   case okf of
+                     Ok f -> (fmap . fmap) f (runConversion ma conn)
+                     Errors errs -> return (Errors errs)
+
+instance Alternative Conversion where
+   empty     = Conversion $ \_conn -> pure empty
+   ma <|> mb = Conversion $ \conn -> do
+                   oka <- runConversion ma conn
+                   case oka of
+                     Ok _     -> return oka
+                     Errors _ -> (oka <|>) <$> runConversion mb conn
+
+instance Monad Conversion where
+   return a = Conversion $ \_conn -> return (return a)
+   m >>= f = Conversion $ \conn -> do
+                 oka <- runConversion m conn
+                 case oka of
+                   Ok a -> runConversion (f a) conn
+                   Errors err -> return (Errors err)
+
+instance MonadPlus Conversion where
+   mzero = empty
+   mplus = (<|>)
+
+conversionMap :: (Ok a -> Ok b) -> Conversion a -> Conversion b
+conversionMap f m = Conversion $ \conn -> f <$> runConversion m conn
+
+conversionError :: Exception err => err -> Conversion a
+conversionError err = Conversion $ \_ -> return (Errors [SomeException err])
+
+newTempName :: Connection -> IO Query
+newTempName Connection{..} = do
+    !n <- atomicModifyIORef connectionTempNameCounter
+          (\n -> let !n' = n+1 in (n', n'))
+    return $! Query $ B8.pack $ "temp" ++ show n
+
+-- FIXME?  What error should getNotification and getCopyData throw?
+fdError :: ByteString -> IOError
+fdError funcName = IOError {
+                     ioe_handle      = Nothing,
+                     ioe_type        = ResourceVanished,
+                     ioe_location    = B8.unpack funcName,
+                     ioe_description = "failed to fetch file descriptor",
+                     ioe_errno       = Nothing,
+                     ioe_filename    = Nothing
+                   }
