@@ -8,7 +8,6 @@
 -- License:     BSD3
 -- Maintainer:  Leon P Smith <leon@melding-monads.com>
 -- Stability:   experimental
--- Portability: portable
 --
 -- Basic types.
 --
@@ -23,6 +22,7 @@ module Database.PostgreSQL.Simple.Types
     , Query(..)
     , Oid(..)
     , (:.)(..)
+    , Savepoint(..)
     ) where
 
 import Blaze.ByteString.Builder (toByteString)
@@ -79,6 +79,7 @@ instance Monoid Query where
     mempty = Query B.empty
     mappend (Query a) (Query b) = Query (B.append a b)
     {-# INLINE mappend #-}
+    mconcat xs = Query (B.concat (map fromQuery xs))
 
 -- | A single-value \"collection\".
 --
@@ -130,3 +131,6 @@ newtype Binary a = Binary a
 data h :. t = h :. t deriving (Eq,Ord,Show,Read,Typeable)
 
 infixr 3 :.
+
+newtype Savepoint = Savepoint Query
+    deriving (Eq, Ord, Show, Read, Typeable)
